# -*- coding: utf-8 -*-

import os
os.chdir("D:/_Travaux/university/Cours_Universite/Supports_de_cours/Informatique/Python/Slides/exemples/K")

#importing the Pandas library
#reading the data file with read.table()
import pandas
cars = pandas.read_table("vehicules_1.txt",sep="\t",header=0,index_col=0)

#dimensions
print(cars.shape)

#number of instances
n = cars.shape[0]

#number of input variables for the regression
p = cars.shape[1] - 1

#list of variables and their types
print(cars.dtypes)

#index of the cars
print(cars.index)

##########################################
# Regression and inspecting the results  #
##########################################

#regression with a formula (as in R)
import statsmodels.formula.api as smf

#instantiation
reg = smf.ols('conso ~ cylindree + puissance + poids', data = cars)

#members of reg
print(dir(reg))

#launching the modelling process
res = reg.fit()

#members of the object provided by the modelling
print(dir(res))

#detailed results
print(res.summary())

#estimated coefficients
print(res.params)

#coefficient of determination R2
print(res.rsquared)

#calculating the F-statistic
F = res.mse_model / res.mse_resid
print(F)

#F provided by the res object
print(res.fvalue)

#test for a linear combination of coefficients
#all the slopes are zero
import numpy as np

#matrix R
R = np.array([[0,1,0,0],[0,0,1,0],[0,0,0,1]])

#F-statistic
print(res.f_test(R))

################################
# Diagnostic of the regression #
################################

### normality of the residuals

#Jarque-Bera normality test
import statsmodels.api as sm
JB, JBpv,skw,kurt = sm.stats.stattools.jarque_bera(res.resid)
print(JB,JBpv,skw,kurt)

#normal probability plot
sm.qqplot(res.resid)

### detection of influential points

#object for measuring influence
infl = res.get_influence()

#members
print(dir(infl))

#leverage
print(infl.hat_matrix_diag)

#internally studentized residuals
print(infl.resid_studentized_internal)

#checking with the formula in the course
import numpy as np
residus = res.resid.as_matrix()
leviers = infl.hat_matrix_diag #leviers
sigma_err = np.sqrt(res.scale) #ec.type.erreur
res_stds = residus/(sigma_err*np.sqrt(1.0-leviers))
print(res_stds)

#externally studentized residuals
print(infl.resid_studentized_external)

#checking with the formula
res_studs = res_stds*np.sqrt((n-p-2)/(n-p-1-res_stds**2))
print(res_studs)

#graphical representation of the influences
sm.graphics.influence_plot(res)

#detection based on the leverage and the extrenally studentized residuals

#threshold value for the leverage
seuil_levier = 2*(p+1)/n
print(seuil_levier)

#identification
atyp_levier = leviers > seuil_levier
print(atyp_levier)

#which cars?
print(cars.index[atyp_levier],leviers[atyp_levier])

#threshold value for externally studentized residuals
import scipy
seuil_stud = scipy.stats.t.ppf(0.975,df=n-p-2)
print(seuil_stud)

#detection - val. abs > threshold
atyp_stud = np.abs(res_studs) > seuil_stud

#which cars?
print(cars.index[atyp_stud],res_studs[atyp_stud])

#identification with at least one of the criteria
pbm_infl = np.logical_or(atyp_levier,atyp_stud)
print(cars.index[pbm_infl])

#presentation in a raw table format
print(infl.summary_frame().filter(["hat_diag","student_resid","dffits","cooks_d"]))

###############################
# Multicollinearity problem   #
###############################

#input variables, numpy matrix format
cars_exog = cars[['cylindree','puissance','poids']].as_matrix()

#correlation matrix with scipy
import scipy
mc = scipy.corrcoef(cars_exog,rowvar=0)
print(mc)

#Klein's rule
mc2 = mc**2
print(mc2)

#VIF - variance inflation factor
vif = np.linalg.inv(mc)
print(vif)


###############################################
# Punctual prediction and prediction interval #
###############################################

#loading the second data file
cars2 = pandas.read_table("vehicules_2.txt",sep="\t",header=0,index_col=0)

#number of instances in the second data file
n_pred = cars2.shape[0]

#cars in the second data file
print(cars2)

#input variables
cars2_exog = cars2[['cylindree','puissance','poids']]

#add a constant to the matrix of input variables
cars2_exog = sm.add_constant(cars2_exog)
print(cars2_exog)

#punctual prediction - reg is the regression object
pred_conso = reg.predict(res.params,cars2_exog)
print(pred_conso)

#comparison observed vs. predicted
import matplotlib.pyplot as plt
plt.scatter(cars2['conso'],pred_conso)
plt.plot(np.arange(5,23),np.arange(5,23))
plt.xlabel('Observed')
plt.ylabel('Predicted')
plt.xlim(5,22)
plt.ylim(5,22)
plt.show()

#(X'X)^-1 matrix from the regression
inv_xtx = reg.normalized_cov_params
print(inv_xtx)

#transformation in a matrix type of the input variables (second data file)
X_pred = cars2_exog.as_matrix()

#square of the standard error of the prediction
#initialization
var_err = np.zeros((n_pred,))
#for each individual to process
for i in range(n_pred):
    #description of the individual
    tmp = X_pred[i,:]
    #matrix product
    pm = np.dot(np.dot(tmp,inv_xtx),np.transpose(tmp))
    #squared values
    var_err[i] = res.scale * (1 + pm)
#
print(var_err)    
    
#quantile of the Student distribution (0.975)
qt = scipy.stats.t.ppf(0.975,df=n-p-1)

#lower bound of the prediction interval
yb = pred_conso - qt * np.sqrt(var_err)
print(yb)

#upper bound of the prediction interval
yh = pred_conso + qt * np.sqrt(var_err)
print(yh)

#matrix with the various values (lower bound, observed, upper bound)
a = np.resize(yb,new_shape=(n_pred,1))
y_obs = cars2['conso']
a = np.append(a,np.resize(y_obs,new_shape=(n_pred,1)),axis=1)
a = np.append(a,np.resize(yh,new_shape=(n_pred,1)),axis=1)
print(a)

#transforming in a data frame object to obtain better displaying
df = pandas.DataFrame(a)
df.index = cars2.index
df.columns = ['B.Basse','Y.Obs','B.Haute']
print(df)